/* ---------- serial.h --------------- */

/*
 * Serial Port Definitions
 */

extern char *next_serial_in,
            *next_serial_out;

/* ----------- serial prototypes ----------- */
void init_serial_port(unsigned long int, int, int, int);
int read_serial(void);
int write_serial(int);
void clear_serial_queue(void);
void restore_serial_int(void);

#define TRUE    1
#define FALSE   0

/* ----------------- serial macros ------------------- */
#define serial_stat() (inp(LINE_STATUS))
#define input_char_ready() (next_serial_in != next_serial_out)
#define XON      17
#define XOFF     19

/* --------------- serial interrupt values ------------------ */
#define PIC01       0x21 /* 8259 Programmable Interrupt Controller */
#define PIC00       0x20 /*  "      "              "        "      */
#define EOI         0x20 /* End of Interrupt command               */
#define TIMER       0x1c /* PC timer interrupt vector              */

/* --------------- line status register values ------------- */
#define XMIT_DATA_READY    0x20

/* ------------ modem control register values -------------- */
#define DTR      1
#define RTS      2
#define OUT2     8

/* ----------- ASCII serial control characters ------------- */
#define SOH      0x01
#define STX      0x02
#define ETX      0x03
#define EOT      0x04
#define ACK      0x06
#define DLE      0x10
#define NAK      0x15
#define CAN      0x18

/* ----------- interrupt enable register signals ----------- */
#define DATA_READY 1

/* ------------- serial input interrupt buffer ------------- */
#define SERIAL_BUFFER_SIZE  1024
#define SAFETY_LEVEL        (SERIAL_BUFFER_SIZE / 4)
#define THRES_HOLD          (SAFETY_LEVEL * 3)
