/* XModem.c */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <ctype.h>
#include <mem.h>     /* Used only for send xmodem */
#include <dos.h>
#include "timer.h"
#include "serial.h"
#include "modem.h"
#include "keys.h"

#define RETRIES     12
#define CRC_TRIES    2
#define PADCHAR   0x1A /* Used only for send xmodem */
#define TRUE         1
#define FALSE        0

#define CRC        'C'

/* External Data */

extern int TIME_OUT;
extern int xonxoff_enabled;

/* Local Data */

static int tries;      /* Retry Counter */
static unsigned char bf [130];  /* I/O Buffer  */
FILE *fp;

/* Prototypes */

int receive_protocol_xmodem (void);
int receive_xmodem (void);
int send_protocol_xmodem (void);
int send_xmodem (void);
static void receive_error (int, int);
unsigned compute_crc (char *, int);
int receive_filename (char *);

/* Error Messages */

static char *errs[] = {
  "Timed Out         ",
  "SOH Invalido      ",
  "Bloco # Invalido  ",
  "CRC Invalido      "
};

int receive_protocol_xmodem (void)
{
  char filename [15];
  int estado;

  write_xserial (ACK);

  if (receive_filename (filename) == TRUE) {
    cprintf ("\n\r  Receber arquivo: %s.\n\r\n\r", filename);
    cprintf ("  Pressione <ESC> para cancelar.\n\r\n\r");
    if ((fp = fopen(filename, "w+b")) != NULL) {
      write_xserial (ACK);
      estado = receive_xmodem ();
    }
    else {
      write_xserial ('!'); /* Abortar Transmissao */
      cprintf("\r\n  Erro de abertura do '%s'.\n\r", filename);
      return FALSE;
    }
  }
  else
    return FALSE;
  fclose (fp);
  return estado;
}

/* receive with xmodem protocol */

int receive_xmodem (void)
{
  unsigned int blck_no, blk = 0, numblk  = 1, neg_blkno;
  int i,
      soh    = 0,
      crc_in = TRUE,
      fst    = TRUE,
      pula   = 0;
  unsigned chksum,
           cs,
           cs1;

  long endoffile;

  xonxoff_enabled = FALSE;
  /* - send Cs then NAKs until the sender starts sending - */
  tries = 0;
  TIME_OUT = 6;
  while (soh != SOH && tries < RETRIES)
    {
      crc_in = (tries++ < CRC_TRIES);
      write_xserial(crc_in ? CRC : NAK);
      soh = read_xserial();
      if (!timed_out() && soh != SOH)
        sleep(6);
    }
  while (tries < RETRIES)
    {
      if (timed_out())
        receive_error(0, NAK);
      /* -- Receive the data and build the file -- */
      cprintf("\r  Recebendo bloco: %d.", numblk);
      if (!fst)
        {
          TIME_OUT = 10;
          soh = read_xserial();
          if (timed_out())
            continue;
          if (soh == CAN)
            break;
          if (soh == EOT)
            {
              write_xserial(ACK);
              break;
            }
        }
      fst = FALSE;
      TIME_OUT = 5;
      blck_no  = read_xserial() & 0xFF;    /* block number   */
      neg_blkno = read_xserial() & 0xFF;  /* 1's complement */
      chksum = 0;
      /* ---- data block ----- */
      for (i = 0; i < 129; i++)
	{
          *(bf + i) = read_xserial();
          if (timed_out())
            break;
          chksum = (chksum + (*(bf + i)) & 0xFF) & 0xFF;
	}
      pula = 0;
      if (timed_out())
        continue;
      /* ---- checksum or crc from sender ---- */
      cs = read_xserial() & 0xFF;
      if (crc_in)
        {
          cs1 = read_xserial() & 0xFF;
          cs = (cs << 8) + cs1;
        }
      if (timed_out())
        continue;
      if (soh != SOH)
        {       /* check the SOH */
          receive_error(1, NAK);
          continue;
        }
      /* --- same as previous block number? --- */
      if (blck_no == blk) {
	pula = 1;
	/* fseek(fp, -128L, 1); */
      }
        /* --- no, next sequential block number? --- */
      else
        if (blck_no != ((blk + 1) & 0xFF))
          {
            receive_error(2, CAN);
            break;
          }
      blk = blck_no;
      numblk++;
      /* --- test the block # 1s complement --- */
      if ((neg_blkno & 0xFF) != (~blk & 0xFF))
        {
          receive_error(2, NAK);
          continue;
        }
      if (crc_in)
        chksum = compute_crc(bf, 130);
      /* --- test chksum or crc vs one sent --- */
      if (cs != chksum)
        {
          receive_error(6, NAK);
	  continue;
        }
      soh = blck_no = neg_blkno = cs = 0;
      tries = 0;
      /* --- write the block to disk --- */
      endoffile = (long) bf [128];
      /* fwrite (bf, 1, (int) bf [128], fp); */
      if (pula == 0)
	grave (fp, bf, endoffile);
      if (kbhit())
        if (getch() == ESC)
          {
            write_xserial(CAN);
            break;
          }
      write_xserial(ACK);
    }
  if (soh == EOT)
    cprintf("\n\r\n\r  Transmisso completa.\n\r");
  else
    cprintf("\n\r\n\r  Transmisso abortada.\n\r");
  xonxoff_enabled = TRUE;
}

static void receive_error (erno, rtn)
int erno,
    rtn;
{
  ++tries;
  if (TIME_OUT == 1)
    cprintf("\r\n  %s (tentativa %d)\n\r", errs [erno], tries);
  write_xserial(rtn);
}

int send_xmodem (void)
{
  int i,
  chksum,
  ln,
  end_of_file = FALSE,
  ans = 0,
  crc_out = 0;
  unsigned crc;
  unsigned int blck_no = 1;
  long lido,
  repetir = 0;

  int wait_response = 0;

  unsigned endoffile;

  xonxoff_enabled = FALSE;
  tries = 0;

  /* wait for the go-ahead from the receiver */

  TIME_OUT = 6;
  while (tries++ < RETRIES && crc_out != NAK && crc_out != CRC)
    crc_out = read_xserial();
  TIME_OUT = 10;

  tries = 0;
  /*  send the file to the receiver ----------- */

  while (tries < RETRIES && !end_of_file && ans != CAN /*&& !timed_out()*/)
    {
      /* ---- read the next data block ----- */
      if (repetir == 0) {
	setmem(bf, 0x80, PADCHAR);
	lido = 128l;
	end_of_file = le (fp, bf, &lido);
	ln = (int) lido;
	if (ln == 0)
	  break;
      }
      else
	repetir = 0;
      chksum = 0;
      cprintf("\r  Transmitindo bloco %d.", blck_no);
      if (kbhit())
        if (getch() == ESC)
          {
            write_xserial(CAN);
            break;
          }
      write_xserial(SOH);       /* SOH           */
      write_xserial(blck_no);   /* block number  */
      write_xserial(~blck_no);  /* 1s complement */
      /* ------- send the data block ------ */
      bf [128] = lido;
      for (i = 0; i < 0x81; i++)
        {
          write_xserial(bf[i]);
          chksum += bf[i];     /* checksum calculation */
	}
      /* -- send error-correcting value (chksum or crc) -- */
      if (crc_out == NAK)
        write_xserial(chksum & 0xFF);
      else
        {
          crc = compute_crc(bf, 130);
          write_xserial((crc >> 8) & 0xFF);
          write_xserial(crc & 0xFF);
	}

      /* Read ACK, NAK, or CAN from receiver */

      ans = read_xserial();
      if (timed_out()) {
	do {
	  cprintf ("\n\r  Esperando resposta %d.\n\r", ++wait_response);
	  ans = read_xserial();
	} while (timed_out() && wait_response < 5);
	wait_response = 0;
      }
      if (ans == ACK)
        {
          blck_no++;
          tries = 0;
        }
      if (ans == NAK)
	{
          end_of_file = FALSE;
	  cprintf("\r\n  Tentativa %d \n\r", ++tries);

	  /* Position to previous block */
	     repetir = 1;
	 /* if (fseek(fp, -128L, 1) == -1) {
	    cprintf ("\n\r  Erro de posicionmento de ponteiro.\n\r");
	    fseek(fp, 0L, 0);
	  } */
        }
    } /* while (tries < ... */
  if (end_of_file)
    {
      write_xserial(EOT);                 /* send the EOT    */
      read_xserial();                     /* wait for an ACK */
      cprintf("\n\r\n\r  Transferncia completada.\n\r");
    }
  else
    cprintf("\n\r\n\r  Transferncia abortada.\n\r");
  xonxoff_enabled = TRUE;
}

int send_protocol_xmodem (void)
{
  int estado;
  char filename [15];

  write_xserial (ACK);

  if (receive_filename (filename) == TRUE)
    if ((fp = fopen(filename, "r+b")) != NULL) {
      cprintf ("\n\r  Transmitir arquivo %s.\n\r\n\r", filename);
      cprintf ("  Pressione <ESC> para cancelar.\n\r\n\r");
      write_xserial (ACK);
      estado = send_xmodem ();
    }
    else {
      cprintf("\r\n  Erro de abertura. Arquivo '%s'.\n\r", filename);
      write_xserial ('!'); /* Abortar transmissao */;
      return FALSE;
    }
  else
    return FALSE;
  fclose (fp);
  return estado;
}

unsigned compute_crc (bf, len)
char *bf;
int  len;
{
int i;

  long crc = 0;
  while (len--)
    {
      crc |= (*bf++) & 0xFF;
      for (i = 0; i < 8; i++)
        {
          crc <<= 1;
	  if (crc & 0x1000000L)
            crc ^= 0x102100L;
        }
    }
  return (unsigned) (crc >> 8);
}

int receive_filename (char *filename)
{
  int tentativa = 0, estado = TRUE;

  do { /* Three times or until success */
    unsigned filecrc = 0; /* given by the host */
    int posi = -1;        /* next position into the string */

    tentativa++;

    do { /* Until get the file name or failure */
      filename [++posi] = read_xserial ();
      if ((int) filename [posi] == FALSE)
	estado = FALSE;
    } while (filename [posi] != '*' && posi < 13 && estado != FALSE);

    if (posi >= 13) /* failure */
      estado = FALSE;

    if (estado == TRUE) {
      int ind;
      short file_crc [2];

      filename [posi] = (char) 0; /* Prepare the file name */

      for (ind = 0; ind < 2; ind++) {
	file_crc [ind] = read_xserial ();
	if (file_crc [ind] == FALSE)
	  estado = FALSE;
      }

      if (estado == TRUE) {
	filecrc = file_crc [0] << 8;
	filecrc |= (unsigned) file_crc [1] & 0xff;
	if (filecrc != compute_crc (filename, strlen (filename))) {
	  estado = FALSE;
	}
      }
    }
    if (estado == FALSE) {
      clear_serial_queue ();
      write_xserial (NAK);
    }

  } while (estado != TRUE && tentativa < 3);

  if (estado == FALSE)
    return FALSE;

  return TRUE;
}
